import java.awt.*;
import java.applet.*;

//	System.out.println(oldx + "," + oldy + "   " + tx + "," + ty);


class ScrollableCanvas extends Canvas {
  public Image image;
  public int tx = 0,ty = 0;
  int WIDTH, HEIGHT, SIZE;
  Graphics osGraphics = null, TileCanvas = null;
  Image osImage = null;
  byte pixels[];

public ScrollableCanvas() {
}

public void addNotify() {
  super.addNotify();
  TileCanvas=getGraphics();
}

public void reset(byte pixels[]) {
  this.pixels=pixels;
  updateImage();
  redraw();
}  

public void reset(int w, int h, int s, byte pixels[]) {
  WIDTH=w; HEIGHT=h; SIZE=s;
  osImage = this.createImage(WIDTH * SIZE+1, HEIGHT * SIZE+1);
  osGraphics = osImage.getGraphics();

  osGraphics.setColor(Color.gray);
  for (int y = 0 ; y < HEIGHT ; y++)
    for (int x = 0 ; x < WIDTH ; x++)
      osGraphics.drawRect(x * SIZE, y * SIZE, SIZE, SIZE);

  this.pixels=pixels;
  updateImage();
  paint(TileCanvas);
}

public void updateImage() {
	int setColor;
  byte osColor=-1;
  int index=0;
  for (int y = 0 ; y < HEIGHT ; y++) {
    for (int x = 0 ; x < WIDTH ; x++) {
      if (pixels[index] != osColor) {
	      setColor = pixels[index];
        if (setColor < 0 ) setColor += 256;
	      osGraphics.setColor(ColorSet.colors[setColor]);
	      osColor = pixels[index];
      }
      osGraphics.fillRect(x * SIZE+1, y * SIZE+1, SIZE-1, SIZE-1);
      index++;
    }
  }

}


public void paint(Graphics g) {
  updateImage();
  g.drawImage(osImage, 0-tx, 0-ty, null);
}

public void paint() {
  updateImage();
  TileCanvas.drawImage(osImage, 0-tx, 0-ty, null);

}

public void redraw() {
  TileCanvas.drawImage(osImage, 0-tx, 0-ty, null);
}

public void plotPoint(int x,int y, int color, boolean mode) {
  //System.out.println("plotting (" + x + "," + y +":"+color+")");
  if (x >= 0 && x < WIDTH && y >=0 && y < HEIGHT) {
    if (mode) {
      osGraphics.setColor(ColorSet.colors[color]);
      osGraphics.fillRect(x * SIZE+1, y * SIZE+1, SIZE-1, SIZE-1);
    }
    else {
      TileCanvas.setColor(ColorSet.colors[color]);
      TileCanvas.fillRect(x * SIZE-tx+1, y * SIZE-ty+1, SIZE-1, SIZE-1);
    }
  }
}

private Rectangle boundsBox(int x1, int y1, int x2, int y2) {
  int sx, sy, ex, ey;
  if (x1 < x2) {
    sx=x1;
    ex=x2;
  }
  else {
    sx=x2;
    ex=x1;
  }
  if (y1 < y2) {
    sy=y1;
    ey=y2;
  }
  else {
    sy=y2;
    ey=y1;
  }
  sx=(sx >= 0 && sx <WIDTH) ? sx :
	(sx < 0 ? 0 : WIDTH-1);
  sy=(sy >= 0 && sy <HEIGHT) ? sy :
	(sy < 0 ? 0 : HEIGHT-1);
  ex=(ex >= 0 && ex <WIDTH) ? ex :
	(ex < 0 ? 0 : WIDTH-1);
  ey=(ey >= 0 && ey <HEIGHT) ? ey :
	(ey < 0 ? 0 : HEIGHT-1);
  return new Rectangle(sx,sy,ex-sx,ey-sy);
}

public void captureBox(int x1, int y1, int x2, int y2, Color color) {
  Rectangle r=boundsBox(x1,y1,x2,y2);
  TileCanvas.setColor(color);
  TileCanvas.drawRect(r.x*SIZE,r.y*SIZE,
	     (r.width+1)*SIZE,
	     (r.height+1)*SIZE);
}


public Dimension minimumSize() {
  return new Dimension(SIZE*2, SIZE*2);
}
  
public Dimension preferredSize() {
  return new Dimension(WIDTH*SIZE, HEIGHT*SIZE);
}
}

public final class TilePanel extends Panel  {

  protected int WIDTH = 32;
  protected int HEIGHT = 32;
  protected int SIZE = 5;
  
  public byte pixels[];

  //  int BORDER=5;
  int mode=0;
  int oldx, oldy, firstx, firsty, lastx, lasty, diag, tdiag, dx, dy;
  boolean mouse_in_applet=false;
  UserMode user;
  
  Scrollbar vert;
  Scrollbar horz;
  ScrollableCanvas canvas;
  
public TilePanel(UserMode u) {
  super();
  user=u;
  canvas = new ScrollableCanvas();
  horz = new Scrollbar(Scrollbar.HORIZONTAL);
  vert = new Scrollbar(Scrollbar.VERTICAL);

  setLayout(new BorderLayout());
  add("Center", canvas);
  add("East", vert);
  add("South", horz);
  
}

public void addNotify() {
  super.addNotify();
  reset();
}

public void paint(Graphics g) {
  Dimension d = this.size();
  g.setColor(Color.lightGray);
  g.draw3DRect(0, 0, d.width - 1, d.height - 1, false);
  validate();
  resizeHorz();
  resizeVert();
  canvas.repaint();
}

public Dimension preferredSize() {
  return(new Dimension(WIDTH * SIZE, HEIGHT * SIZE));
}

public Dimension minimumSize() {
  return(new Dimension(WIDTH * SIZE, HEIGHT * SIZE));
}


  /* ---------------main code----------------- */


public void reset() {
  WIDTH=user.IconWidth;
  HEIGHT=user.IconHeight;
  SIZE=user.magni;
  pixels = user.Icons[user.currIcon];
  canvas.reset(WIDTH,HEIGHT,SIZE,pixels);
  validate();
  resizeHorz();
  resizeVert();
  repaint();
}


public void resync() {
  pixels = user.Icons[user.currIcon];
  canvas.reset(pixels);
}

private void resizeHorz() {
  int canvasWidth = canvas.size().width;
   
  if (canvasWidth <= 0) {
    System.out.println("Canvas has no width; can't resize scrollbar");
    return;
  }
  
  //Shift everything to the right if we're displaying empty space
  //on the right side.
  if ((canvas.tx + canvasWidth) > WIDTH*SIZE) {
    int newtx = WIDTH*SIZE - canvasWidth;
    if (newtx < 0) {
      newtx = 0;
    }
    canvas.tx = newtx;
  }
  
  horz.setValues(//draw the part of the image that starts at this x:
		 canvas.tx, 
		 //amount to scroll for a "page":
		 (int)(canvasWidth), 
		 //minimum image x to specify:
		 0,
		 //maximum image x to specify:
		 WIDTH*SIZE); // XXX -canvasWidth for windows?
  //"visible" arg to setValues() has no effect after scrollbar is visible.
  horz.setPageIncrement((int)(canvasWidth * 0.9));
  return;
}
  
  //Don't call this until the canvas size is valid.
private  void resizeVert() {
  int canvasHeight = canvas.size().height;
  
  if (canvasHeight <= 0) {
    System.out.println("Canvas has no height; can't resize scrollbar");
    return;
  }
  
  //Shift everything downward if we're displaying empty space
  //on the bottom.
  if ((canvas.ty + canvasHeight) > HEIGHT*SIZE) {
    int newty = HEIGHT*SIZE - canvasHeight;
    if (newty < 0) {
      newty = 0;
    }
    canvas.ty = newty;
  }
  
  vert.setValues(//initially draw part of image starting at this y:
		 canvas.ty, 
		 //visible arg--amount to scroll for a "page":
		 (int)(canvasHeight), 
		 //minimum image y to specify:
		 0,
		 //maximum image y to specify:
		 HEIGHT*SIZE); // XXX -canvasHeight for windows?
  
  //"visible" arg to setValues() has no effect after scrollbar is visible.
  vert.setPageIncrement((int)(canvasHeight * 0.9));
  return;
}
  

 // events
public boolean handleEvent(Event evt) {
  int tx = ((evt.x+canvas.tx) / SIZE);
  int ty = ((evt.y+canvas.ty) / SIZE);
  boolean handled=false;
  switch(evt.id) {
  case Event.MOUSE_MOVE:
    if ((oldx!=tx || oldy!=ty) && mouse_in_applet) {
      if (user.pastemode) {
	canvas.captureBox( 
		   oldx-user.clipboard.width/2, 
		   oldy-user.clipboard.height/2,
		   oldx-user.clipboard.width/2+user.clipboard.width-1, 
		   oldy-user.clipboard.height/2+user.clipboard.height-1,
		   Color.gray);
	canvas.captureBox(
		   tx-user.clipboard.width/2, 
		   ty-user.clipboard.height/2,
		   tx-user.clipboard.width/2+user.clipboard.width-1, 
		   ty-user.clipboard.height/2+user.clipboard.height-1,
		   Color.white);
      //      parent.showStatus("x:" + tx + "  y:" + ty);
      }
      oldx=tx;
      oldy=ty;
    }
    break;
  case Event.MOUSE_DRAG:
    if ((oldx!=tx || oldy!=ty) && !user.pastemode) {
      if (evt.modifiers == 0 )       {
        switch(user.mode) {
        case 1:
	        plotLine(oldx,oldy,tx,ty, user.color,true);
	        canvas.redraw();
	        user.tileicon.putIcon();
	        break;
        case 2:
	        plotPoint(tx,ty, user.color, true);
	        canvas.redraw();
	        user.tileicon.putIcon();
	        break;
        case 3:
        case 4:
	        canvas.redraw();
	        plotLine(firstx,firsty,tx,ty, user.color, false);
	        break;
        case 5:
	        canvas.redraw();
	        plotRect(firstx,firsty,tx,ty, user.color, false, false);
	        break;
        case 6:
	        canvas.redraw();
	        plotEllipse(firstx,firsty,tx,ty, user.color, false, false);
	        break;
        case 7:
  	      dx=tx-firstx;
	        dy=ty-firsty;
	        tdiag = (dx > 0 ? dx : -dx) > (dy > 0 ? dy : -dy) ? 
	          (dx > 0 ? dx : -dx) : (dy > 0 ? dy : -dy);
  	      if (tdiag!=diag) {
	          canvas.redraw();
	          diag=tdiag;
	          plotEllipse(firstx, firsty,
	  	      firstx + (tx < firstx ? -diag : diag), 
		        firsty + (ty < firsty ? -diag : diag),
		        user.color, false,false);
  	      }
	        break;
        case 9:
	        canvas.captureBox(firstx, firsty, oldx, oldy, Color.gray);
	        canvas.captureBox(firstx, firsty, tx, ty, Color.white);
	        break;
        case 10:
          canvas.redraw();
	        plotRect(firstx,firsty,tx,ty, user.color, true, false);
	        break;
        case 11:
          canvas.redraw();
	        plotEllipse(firstx,firsty,tx,ty, user.color, true, false);
	        break;
        case 12:
  	      dx=tx-firstx;
	        dy=ty-firsty;
	        tdiag = (dx > 0 ? dx : -dx) > (dy > 0 ? dy : -dy) ? 
	        (dx > 0 ? dx : -dx) : (dy > 0 ? dy : -dy);
	        if (tdiag!=diag) {
	          canvas.redraw();
	          diag=tdiag;
	          plotEllipse(firstx, firsty,
	  	      firstx + (tx < firstx ? -diag : diag), 
	  	      firsty + (ty < firsty ? -diag : diag),
  		      user.color, true,false);    
	        }
	        break;
 
        }//end case
        
      }//end if
    }
    oldx=tx;
    oldy=ty;
    break;
    
  case Event.MOUSE_UP:
    //    if (mouse_in_applet)
    //  parent.showStatus("x:" + tx + "  y:" + ty);
    if (user.pastemode) {
      user.pastemode=false;
    }
    else {
      if (evt.modifiers == 0 )       {
        switch(user.mode) {
        case 3:
        case 4:
          user.undobuf=new IconBuf(pixels,WIDTH,HEIGHT,firstx,firsty,tx,ty);
          plotLine(firstx,firsty,tx,ty, user.color, true);
          canvas.redraw();
          user.tileicon.putIcon();
          lastx=tx;
          lasty=ty;
          if (user.mode==4 && tx==user.firstx && ty==user.firsty)
	          user.newmode=true;
          break;
        case 5:
          user.undobuf=new IconBuf(pixels,WIDTH,HEIGHT,firstx,firsty,tx,ty);
          plotRect(firstx,firsty,tx,ty, user.color, false, true);
          canvas.redraw();
          user.tileicon.putIcon();
          break;
        case 6:
          user.undobuf=new IconBuf(pixels,WIDTH,HEIGHT,firstx,firsty,tx,ty);
          plotEllipse(firstx,firsty,tx,ty,user.color, false, true);
          canvas.redraw();
          user.tileicon.putIcon();
          break;
        case 7:
          user.undobuf=new IconBuf(pixels,WIDTH,HEIGHT,firstx,firsty,tx,ty);
          dx=tx-firstx;
          dy=ty-firsty;
          diag = (dx > 0 ? dx : -dx) > (dy > 0 ? dy : -dy) ? 
	          (dx > 0 ? dx : -dx) : (dy > 0 ? dy : -dy);
          plotEllipse(firstx, firsty,
		      firstx + (tx < firstx ? -diag : diag), 
		      firsty + (ty < firsty ? -diag : diag),
		      user.color, false, true);
          canvas.redraw();
          user.tileicon.putIcon();
          break;
        case 9:
          user.undobuf=new IconBuf(pixels,WIDTH,HEIGHT,firstx,firsty,tx,ty);
          user.captx1=(firstx >= 0 && firstx <WIDTH) ? firstx :
	          (firstx < 0 ? 0 : WIDTH-1);
          user.capty1=(firsty >= 0 && firsty <HEIGHT) ? firsty :
	          (firsty < 0 ? 0 : HEIGHT-1);
          user.captx2=(tx >= 0 && tx <WIDTH) ? tx :
	          (tx < 0 ? 0 : WIDTH-1);
          user.capty2=(ty >= 0 && ty <HEIGHT) ? ty :
	          (ty < 0 ? 0 : HEIGHT-1);
          canvas.captureBox(firstx, firsty, oldx, oldy, Color.gray);
          user.tileicon.putIcon();
          break;
        case 10:
          user.undobuf=new IconBuf(pixels,WIDTH,HEIGHT,firstx,firsty,tx,ty);
          plotRect(firstx,firsty,tx,ty, user.color, true, true);
          canvas.redraw();
          user.tileicon.putIcon();
          break;
        case 11:
          user.undobuf=new IconBuf(pixels,WIDTH,HEIGHT,firstx,firsty,tx,ty);
          plotEllipse(firstx,firsty,tx,ty,user.color, true, true);
          canvas.redraw();
          user.tileicon.putIcon();
          break;
        case 12:
          user.undobuf=new IconBuf(pixels,WIDTH,HEIGHT,firstx,firsty,tx,ty);
          dx=tx-firstx;
          dy=ty-firsty;
          diag = (dx > 0 ? dx : -dx) > (dy > 0 ? dy : -dy) ? 
	          (dx > 0 ? dx : -dx) : (dy > 0 ? dy : -dy);
          plotEllipse(firstx, firsty,
		      firstx + (tx < firstx ? -diag : diag), 
		      firsty + (ty < firsty ? -diag : diag),
		      user.color, true, true);
          canvas.redraw();
          user.tileicon.putIcon();
          break;
        }
      }
    break;
  }
  case Event.MOUSE_DOWN:
    if (user.pastemode) {
      user.undobuf=new IconBuf(pixels,WIDTH,HEIGHT,
			       tx-user.clipboard.width/2,
			       ty-user.clipboard.height/2,
			       tx-user.clipboard.width/2+user.clipboard.width,
			       ty-user.clipboard.height/2+user.clipboard.height);
      user.clipboard.Paste(pixels,WIDTH,HEIGHT,tx-user.clipboard.width/2,
			   ty-user.clipboard.height/2);
      canvas.paint();
      user.tileicon.putIcon();
    }
    else {
      if (evt.modifiers != 0 )       {
      	int setColor = (int)pixels[tx+ty*WIDTH];
      	if (setColor < 0 ) setColor += 256;
      	bitedit.colorArea.remotePush(setColor);
      }
      else {    
        switch (user.mode) {
        case 1:
        case 2:
          user.undobuf=new IconBuf(pixels,WIDTH,HEIGHT,0,0,WIDTH-1,HEIGHT-1);
          plotPoint(tx,ty, user.color, true);
          canvas.paint();
	        user.tileicon.putIcon();
          break;
        case 3:
        case 5:
        case 6:
        case 7:
        case 9:
        case 10:
        case 11:
        case 12:
          diag=0;
          firstx=tx;
          firsty=ty;
          break;
        case 4:
          if (user.newmode) {
	          user.firstx=firstx=tx;
	          user.firsty=firsty=ty;
	          user.newmode=false;
          }
          else {
  	        firstx=lastx;
         	  firsty=lasty;
          }
          break;
        case 8:
          user.undobuf=new IconBuf(pixels,WIDTH,HEIGHT,0,0,WIDTH-1,HEIGHT-1);
          if (tx >=0 && tx < WIDTH && ty >=0 && ty < HEIGHT) {
	          plotFill(tx,ty,user.color);
	          canvas.paint();
	          user.tileicon.putIcon();
        }
        break;
        }
      oldx=tx;
      oldy=ty;
      break;
      }
  }
  
  case Event.MOUSE_ENTER:
    mouse_in_applet=true;
    
    break;
  case Event.MOUSE_EXIT:
    //    parent.showStatus("");
    mouse_in_applet=false;
    break;
  case Event.ACTION_EVENT:
    break;
  case Event.SCROLL_LINE_UP:
  case Event.SCROLL_LINE_DOWN:
  case Event.SCROLL_PAGE_UP:
  case Event.SCROLL_PAGE_DOWN:
  case Event.SCROLL_ABSOLUTE:
    if (evt.target == vert) {
      canvas.ty = ((Integer)evt.arg).intValue();
      canvas.redraw();
    }
    if (evt.target == horz) {
      canvas.tx = ((Integer)evt.arg).intValue();
      canvas.redraw();
    }
    break;
  default:
    break;
  }
  return true;
}


  /* ----------------- graphing functions --------------- */

private Rectangle boundsBox(int x1, int y1, int x2, int y2) {
  int sx, sy, ex, ey;
  if (x1 < x2) {
    sx=x1;
    ex=x2;
  }
  else {
    sx=x2;
    ex=x1;
  }
  if (y1 < y2) {
    sy=y1;
    ey=y2;
  }
  else {
    sy=y2;
    ey=y1;
  }
  sx=(sx >= 0 && sx <WIDTH) ? sx :
	(sx < 0 ? 0 : WIDTH-1);
  sy=(sy >= 0 && sy <HEIGHT) ? sy :
	(sy < 0 ? 0 : HEIGHT-1);
  ex=(ex >= 0 && ex <WIDTH) ? ex :
	(ex < 0 ? 0 : WIDTH-1);
  ey=(ey >= 0 && ey <HEIGHT) ? ey :
	(ey < 0 ? 0 : HEIGHT-1);
  return new Rectangle(sx,sy,ex-sx,ey-sy);
}

public void plotPoint(int x,int y, int color, boolean mode) {
  //  System.out.println("plotting (" + x + "," + y +")");
  if (x >= 0 && x < WIDTH && y >=0 && y < HEIGHT && color!=pixels[x+y*WIDTH]) {
    if (mode) { 
      pixels[x+y*WIDTH]=(byte)color;
    }
    canvas.plotPoint(x, y, color, mode);
  }
}

public void plotLine(int x1, int y1, int x2, int y2, int color, boolean mode) {
  int i, dx, dy, l, m, x_inc, y_inc, err_1, err_2, dx2, dy2, x, y;

  x = x1;
  y = y1;
  dx = x2 - x1;
  dy = y2 - y1;
  x_inc = (dx < 0) ? -1 : 1;
  l = dx >= 0 ? dx : -dx;
  y_inc = (dy < 0) ? -1 : 1;
  m = dy >= 0 ? dy : -dy;
  dx2 = l << 1;
  dy2 = m << 1;
  
  if (l >= m) {
    err_1 = dy2 - l;
    for (i = 0; i < l; i++) {
      plotPoint(x,y,color,mode);
      if (err_1 > 0) {
	y += y_inc;
	err_1 -= dx2;
      }
      err_1 += dy2;
      x += x_inc;
    }
  }
  else if (m >= l) {
    err_1 = dx2 - m;
    for (i = 0; i < m; i++) {
      plotPoint(x,y,color,mode);
      if (err_1 > 0) {
	x += x_inc;
	err_1 -= dy2;
      }
      err_1 += dx2;
      y += y_inc;
    }
  }
  plotPoint(x,y,color,mode);
}

public void plotRect(int x1, int y1, int x2, int y2, int color,
		     boolean fillmode, boolean mode) {
  if (!fillmode) {
    plotLine(x1,y1,x2,y1,color, mode);
    plotLine(x2,y1,x2,y2,color, mode);
    plotLine(x1,y2,x2,y2,color, mode);
    plotLine(x1,y1,x1,y2,color, mode);
  }
  else {
    if (x1 < x2)
      for (int x=x1; x<=x2; x++)
	      plotLine(x,y1,x,y2,color, mode);
    else
      for (int x=x2; x<=x1; x++)
	      plotLine(x,y1,x,y2,color, mode);
    }
  }

public void plotEllipse(int x1, int y1, int x2, int y2, 
			int color, boolean fillmode, boolean mode) {
    int x,y,a2,b2, S, T,a,b,cx,cy;
    a=(x2-x1)/2;
    b=(y2-y1)/2;
    cx=a+x1;
    cy=b+y1;
    a= a < 0 ? -a : a;
    b= b < 0 ? -b : b;
    a2 = a*a;
    b2 = b*b;
    if (b > 0) {
      x = 0;
      y = b;
      S = a2*(1-2*b) + 2*b2;
      T = b2 - 2*a2*(2*b-1);
      if (!fillmode) {
	plotPoint(cx+x,cy+y,color, mode);
	plotPoint(cx-x,cy+y,color, mode);
	plotPoint(cx-x,cy-y,color, mode);
	plotPoint(cx+x,cy-y,color, mode);
      }
      else {
	plotLine(cx-x,cy+y,cx+x,cy+y,color, mode);
	plotLine(cx-x,cy-y,cx+x,cy-y,color, mode);
      }
      do {
        if (S<0) {
	  S += 2*b2*(2*x+3);
	  T += 4*b2*(x+1);
	  x++;
        } else if (T<0) {
	  S += 2*b2*(2*x+3) - 4*a2*(y-1);
	  T += 4*b2*(x+1) - 2*a2*(2*y-3);
	  x++;
	  y--;
        } else {
	  S -= 4*a2*(y-1);
	  T -= 2*a2*(2*y-3);
	  y--;
        }
	if (!fillmode) {
	  plotPoint(cx+x,cy+y,color, mode);
	  plotPoint(cx-x,cy+y,color, mode);
	  plotPoint(cx-x,cy-y,color, mode);
	  plotPoint(cx+x,cy-y,color, mode);
	}
	else {
	  plotLine(cx-x,cy+y,cx+x,cy+y,color, mode);
	  plotLine(cx-x,cy-y,cx+x,cy-y,color, mode);
	}
      } while (y>0);
    }
}

public void plotFill(int x, int y, int color) {
  boolean lastBorder;
  int leftLimit, rightLimit,i;
  int yval=y*WIDTH;
  if (color < 0) color +=256;
  int old = pixels[x+yval];
  if (old == color) {
    /* Nothing to be done */
    return;
  }
  /* Seek left */
  leftLimit = (-1);
  for (i = x; (i >= 0); i--) {
    if (pixels[i+yval] != old) {
      break;
    }
    pixels[i+yval]=(byte)color;
    leftLimit = i;
  }
  if (leftLimit == (-1)) {
    return;
  }
  /* Seek right */
  rightLimit = x;
  for (i = (x+1); (i < WIDTH); i++) {	
    if (pixels[i+yval] != old) {
      break;
    }
    pixels[i+yval]=(byte)color;
    rightLimit = i;
  }
  /* Look at lines above and below and start paints */
  /* Above */
  if (y > 0) {
    lastBorder = true;
    yval=(y-1)*WIDTH;
    for (i = leftLimit; (i <= rightLimit); i++) {
      int c;
      c = pixels[i+yval];
      if (lastBorder) {
	      if (c == old) {	
	        plotFill(i, y-1, color);		
	        lastBorder = false;
	      }
      } else if (c != old) {
	    lastBorder = true;
      }
    }
  }
  /* Below */
  if (y < HEIGHT - 1) {
    lastBorder = true;
    yval=(y+1)*WIDTH;
    for (i = leftLimit; (i <= rightLimit); i++) {
      int c;
      c = pixels[i+yval];
      if (lastBorder) {
	if (c == old) {
	  plotFill(i, y+1, color);		
	  lastBorder = false;
	}
      } else if (c != old) {
	lastBorder = true;
      }
    }
  }
}
  
public void flipRect(int x1, int y1, int x2, int y2,
		     boolean horizontal) {
  int j,k;
  Rectangle r=boundsBox(x1,y1,x2,y2);
  byte temp;
  if (horizontal)
    for (int i=r.y+r.height; i >= r.y; i--) {
      j=r.x; 
      k=r.x+r.width;
      while (j < k) {
	temp=pixels[j+i*WIDTH];
	pixels[j+i*WIDTH]=pixels[k+i*WIDTH];
	pixels[k+i*WIDTH]=temp;
	j++; k--;
      }
    }
  else 
    for (int i=r.x+r.width; i >= r.x; i--) {
      j=r.y; 
      k=r.y+r.height;
      while (j < k) {
	temp=pixels[j*WIDTH+i];
	pixels[i+j*WIDTH]=pixels[i+k*WIDTH];
	pixels[i+k*WIDTH]=temp;
	j++; k--;
      }
    }
}

public void invertRect(int x1, int y1, int x2, int y2) {
  Rectangle r=boundsBox(x1,y1,x2,y2);
  byte temp;
    for (int i=r.y+r.height; i >= r.y; i--) {
      for (int j=r.x;j<=r.x+r.width;j++) {
	pixels[j+i*WIDTH]= (pixels[j+i*WIDTH]==1)?0:(byte)1;
      }
    }
}

}
