// bitedit is a standalone Java application to edit bitmaps, specifically
// for the Pilot.
//
// bitedit is based on javaIcon by Steve Tomljenovic
// http://sthelens.neog.com/~stomljen/
//
// The Pilot modifications are by Ken Shirriff
// http://www.sunlabs.com/~shirriff
import java.awt.*;
import java.awt.image.ImageObserver;
import java.net.URL;
import java.lang.*;
import java.io.*;
import java.applet.Applet;



class ErrorDialog extends Dialog {
  bitedit parent;
 
public ErrorDialog(bitedit parent, String message) {
  super(parent, "Error Dialog",false);
  this.parent = parent;
  
  add("North",new Label(message, Label.CENTER));
  add("South", new Button("OK"));
  resize(200,100);
  parent.disable();
}
  
public boolean handleEvent(Event evt) {
  if (evt.id ==Event.WINDOW_DESTROY) {
    parent.enable();
    parent.requestFocus();
    dispose();
    parent.repaint();
    return(true);
  }
  return super.handleEvent(evt);
}

public boolean action(Event evt, Object obj) {
  if (obj.equals("OK")) {
    parent.enable();
    parent.requestFocus();
    dispose();
    parent.repaint();
    return(true);
  }
  return(false);
}
}

class OptionDialog extends Dialog {

  bitedit parent;
  TextField tf[];
  final String types[]={"Width","Height"};
  String title;
  boolean mode;

public OptionDialog(bitedit parent, String title, boolean mode) {
  super(parent, title, false);
  this.parent = parent;
  this.title=title;
  this.mode=mode;
  Panel p = new Panel();
  p.setLayout(new GridLayout(types.length,2));
  tf=new TextField[types.length];
  
  String text="";
  for (int i=0; i < types.length; i++) {
    switch(i) {
    case 0:
      text=(new Integer(parent.user.IconWidth)).toString();
      break;
    case 1:
      text=(new Integer(parent.user.IconHeight)).toString();
      break;
    case 2:
      text=(new Integer(parent.user.IconNum)).toString();
      break;
    }      
    tf[i] = new TextField(text);
    p.add(new Label(types[i]));
    p.add(tf[i]);
  }
  add("Center", p);
  
  Panel j = new Panel();
  j.setLayout(new FlowLayout());
  j.add(new Button("OK"));
  j.add(new Button("Cancel"));

  add("South",j);

  requestFocus();
  parent.disable();
}
  
public boolean handleEvent(Event evt) {
  if (evt.id==Event.WINDOW_DESTROY) {
    parent.enable();
    parent.requestFocus();
    dispose();
    parent.repaint();
    return(true);
  }
  return(super.handleEvent(evt));
}

public boolean action(Event evt, Object obj) {
  boolean ok=true;
  int w=0,h=0,n=1;
  if (obj.equals("OK")) {
    try {
      w=Integer.parseInt(tf[0].getText());
      h=Integer.parseInt(tf[1].getText());
    } catch (NumberFormatException e) {
      ok=false;
    }
    if (ok) {
      parent.enable();
      parent.requestFocus();
      dispose();
      if (mode)
	parent.user.newTiles(w,h,n,true);
      else
	parent.user.resize(w,h,n);
      parent.user.reset();
      parent.repaint();
      resize(500,500);
      pack();
    }
    return(true);
  }
  if (obj.equals("Cancel")) {
    parent.enable();
    parent.requestFocus();
    dispose();
    parent.repaint();
    return(true);
  }
  return(false);
}
}

public final class bitedit extends Frame {

  UserMode user=new UserMode();
  static boolean inAnApplet=true;

  TilePanel tileArea = null;
  UtilPanel utilArea = null;
  static ColorPanel colorArea = null;
  TileIcon iconArea = null;

  LayoutManager theBoss = new BorderLayout();

  MenuBar mb;
  Menu filemenu,editmenu,optionmenu;
  final String fileitems[]={"New","Open","Save","Quit"};
  final String edititems[]={"Undo","Paste","Cut","Copy","-","Flip Horiz.","Flip Vert.", "Invert"};
  final String optionitems[]={"Resize","Clear Icon","-",
			      "x2","x4","x6","x8","x10"};

  private FileDialog openDialog;
  private FileDialog saveDialog;
  private ErrorDialog errorDialog = null;
  private OptionDialog optionDialog = null;

  String fileName="untitled.jbm";
  
public bitedit() {
  super("bitedit");
  
  mb = new MenuBar();
  filemenu = new Menu("File");
  for (int i=0; i < fileitems.length; i++) 
    filemenu.add(new MenuItem(fileitems[i]));
  mb.add(filemenu);

  editmenu = new Menu("Edit");
  for (int i=0; i < edititems.length; i++) 
    editmenu.add(new MenuItem(edititems[i]));
  mb.add(editmenu);

  optionmenu = new Menu("Option");
  for (int i=0; i < optionitems.length; i++) 
    optionmenu.add(new MenuItem(optionitems[i]));
  mb.add(optionmenu);

  setMenuBar(mb);
  setLayout(theBoss);

  Panel p = new Panel();
  utilArea = new UtilPanel(user);
  user.tilepanel=tileArea = new TilePanel(user);
  user.tileicon=iconArea = new TileIcon(user);
  colorArea = new ColorPanel(user);
  p.setLayout(new BorderLayout());
  p.add("North",utilArea);
  p.add("Center",colorArea);

  add("West",p);
  add("East",iconArea);
  add("Center",tileArea); 

  openDialog = new FileDialog(this, "Open File...", FileDialog.LOAD);
  saveDialog = new FileDialog(this, "Save File...", FileDialog.SAVE);
}

public void showErrorDialog(String message) {
  if (errorDialog != null)
    errorDialog.dispose();
  errorDialog = new ErrorDialog(this, message);
  errorDialog.show();
}

public void showOptionDialog(boolean option) {
  if (optionDialog != null) 
    optionDialog.dispose();
  optionDialog = new OptionDialog(this, option ? "New Tile Set" : "Resize",
				  option);
  optionDialog.pack();
  optionDialog.show();
}

private boolean readBin(String filename) throws Exception {
  DataInputStream is = new DataInputStream(new FileInputStream(filename));
  int w = is.readShort();
  int h = is.readShort();
  int chars_row = is.readShort();
  int fieldflags = is.readShort();
  is.skip(8);
  user.newTiles(w,h,1,false);
  byte p[] = new byte[chars_row];
  for (int i=0; i<h; i++) {
    is.read(p);
    for (int j=0; j<w; j++) {
      user.Icons[0][i*w+j] = ((p[(j>>3)]&(1<<(7-(j&7))))==0)?0:(byte)1;
    }
  }
  return true;
}

private boolean readBmp(String filename) throws Exception {
  DataInputStream is = new DataInputStream(new FileInputStream(filename));
  byte hdr[] = new byte[54];
  is.read(hdr);
  if (hdr[0] != 'B' || hdr[1] != 'M') {
	return false;
  }
  int bfSize = getLong(hdr, 2);
  int biSize = getLong(hdr, 14);
  int biWidth = getLong(hdr, 18);
  int biHeight = getLong(hdr, 22);
  int biPlanes = getShort(hdr, 26);
  int biBitCount = getShort(hdr, 28);
  int biCompression = getLong(hdr, 30);
  int biSizeImage = getLong(hdr, 34);
  int biXPelsPerMeter = getLong(hdr, 38);
  int biYPelsPerMeter = getLong(hdr, 42);
  int biClrUsed = getLong(hdr, 46);
  int biClrImportant = getLong(hdr, 50);

  if (biBitCount != 1) {
    throw new Exception("File is not monochrome");
  }
  byte p[] = new byte[biSizeImage];
  is.skip(40-biSize+8);
  is.read(p);
  user.newTiles(biWidth,biHeight,1,false);
  for (int i=0; i<biHeight; i++) {
    for (int j=0; j<biWidth; j++) {
      user.Icons[0][i*biWidth+j] = testBmpBit(biWidth, biHeight, p, j,
	biHeight-i-1, 32);
    }
  }
  return true;
}

private byte testBmpBit(int cx, int cy, byte p[], int x, int y, int align) {
  int cbRow = (cx + (align-1))& ~(align-1);
  byte b = p[(cbRow>>3)*y+(x>>3)];
  return (b&(1<<(7-(x&7))))==0?0:(byte)1;
}

private int getUByte(byte b[], int off) {
	return ((int)b[off])&0xff;
}

private int getLong(byte b[], int off) {
	return getUByte(b,off)|(getUByte(b,off+1)<<8)|
	  (getUByte(b,off+2)<<16)|(getUByte(b,off+3)<<24);
}

private int getShort(byte b[], int off) {
	return getUByte(b,off)|(getUByte(b,off+1)<<8);
}

private boolean readPbitm(String filename) throws IOException {
  int w=0,h=0;
  DataInputStream is = new DataInputStream(new FileInputStream(filename));
  try {
	  for (h=0; ;h++) {
		String l = is.readLine();
		if (l == null) {
			break;
		}
		if (l.length()>w) {
			w = l.length();
		}
	  }
  } catch (EOFException e) {}
  is.close();
  user.newTiles(w,h,1,false);
  is = new DataInputStream(new FileInputStream(filename));
  for (int i=0; i<h; i++) {
	String l = is.readLine();
	for (int j=0; j<w; j++) {
		if (l.charAt(j) == '#') {
			user.Icons[0][i*w+j] = 1;
		} else if (l.charAt(j) == '.' || l.charAt(j) == '-' ||
		  l.charAt(j) == ' ') {
			user.Icons[0][i*w+j] = 0;
		} else {
			return false;
		}
	}
  }
  is.close();
  return true;
}

private boolean read(String filename) {
  int type, state=0, len=0, index=0, w=0,h=0,n=0;
  InputStream is = null;
  boolean ok = false;
      
  try {
    ok = readBmp(filename);
    if (!ok) {
      ok = readPbitm(filename);
    }
    if (!ok) {
      ok = readBin(filename);
    }
  } catch(Throwable e) {
    showErrorDialog("Can't open \""+filename+"\"");
    return false;
  }
  if (!ok) {
    showErrorDialog("Bad format of \""+filename+"\"");
    return false;
  }
  
  setTitle("bitedit - " + filename );
  return(true);
}
  
private boolean write(String filename) {
  OutputStream os = null;
    
  try {
    if (filename.endsWith(".bin") || filename.endsWith(".BIN")) {
      writeBin(filename);
    } else {
      writePbitm(filename);
    }
  } catch(IOException e) {
    showErrorDialog("Can't write \""+filename+"\"");
    return(false);
  }
  setTitle("bitedit - " + filename );
  return(true);
}

private void writePbitm(String filename) throws IOException {
    OutputStream os = new FileOutputStream(filename);
    byte p[] = new byte[user.IconWidth+1];
    p[user.IconWidth] = '\n';
    for (int i=0; i<user.IconHeight; i++) {
	    for (int j=0; j<user.IconWidth; j++) {
		if (user.Icons[0][i*user.IconWidth+j] != 0) {
			p[j] = '#';
		} else {
			p[j] = '-';
		}
	    }
	    os.write(p);
    }
    os.close();
}

private void writeBin(String filename) throws IOException {
    DataOutputStream os = new DataOutputStream(new FileOutputStream(filename));
    int chars_row = ((user.IconWidth+15)>>4)<<1;
    os.writeShort(user.IconWidth);
    os.writeShort(user.IconHeight);
    os.writeShort(chars_row);
    os.writeShort(0);
    os.writeShort(0);
    os.writeShort(0);
    os.writeShort(0);
    os.writeShort(0);
    for (int i=0; i<user.IconHeight; i++) {
	    byte p[] = new byte[chars_row];
	    for (int j=0; j<user.IconWidth; j++) {
		if (user.Icons[0][i*user.IconWidth+j] != 0) {
		  p[j>>3] |= (1<<(7-(j&7)));
		}
    	    }
	    os.write(p);
    }
    os.close();
}
  
  // Compensate for dialog bug
private String check(String filename) {
  if (filename.endsWith(".*.*")) {
    filename = filename.substring(0, filename.length()-4);
  }
  return(filename);
}
 

public boolean handleEvent(Event evt) {
  if ((evt.id == Event.ACTION_EVENT) 
      && (evt.target instanceof MenuItem)) {
    String label = (String)evt.arg;
    if (label.equals("New")) {
      showOptionDialog(true);
      return true;
    }
    if (label.equals("Open")) {
      // Read in an existing file
      String filename;
      openDialog.show();
      filename = openDialog.getFile();
      if (filename != null) {
	filename = check(filename);
	if (read(openDialog.getDirectory()+filename)) { 
	  fileName = filename;
	  this.requestFocus();
	  resize(500,500);
	  pack();
	  repaint();
	  user.reset();
	}
      }
      return true;
    }
    if (label.equals("Save")) {
      String filename;
      saveDialog.show();
      filename = saveDialog.getFile();
      if (filename != null) {
	filename = check(filename);
	if (write(saveDialog.getDirectory()+filename)) {
	  fileName = filename;
	  this.requestFocus();
	  repaint();
	}
      }
      return true;
    }
    if(label.equals("Quit")) {
      System.exit(0);
    }
    if (label.equals("Undo") && user.undobuf!=null) {
      user.undobuf.Paste(user.tilepanel.pixels,user.tilepanel.WIDTH,
			 user.tilepanel.HEIGHT);
      user.undobuf=null;
      user.tilepanel.resync();
      user.tileicon.putIcon();
      return true;
    }
    if (label.equals("Paste") && user.clipboard!=null) {
      user.pastemode=true;
      return true;
    }
    if (label.equals("Cut")) {
      user.undobuf=user.clipboard=new IconBuf(user.tilepanel.pixels,
					      user.tilepanel.WIDTH,
					      user.tilepanel.HEIGHT,
					      user.captx1,user.capty1,
					      user.captx2,user.capty2);
      user.tilepanel.plotRect(user.captx1,user.capty1,
			      user.captx2,user.capty2,
			      0, true, true);
      user.pastemode=false;
      user.tilepanel.resync();
      user.tileicon.putIcon();
      return true;
    }
    if (label.equals("Copy")) {
      user.clipboard=new IconBuf(user.tilepanel.pixels,
				 user.tilepanel.WIDTH,
				 user.tilepanel.HEIGHT,
				 user.captx1,user.capty1,
				 user.captx2,user.capty2);
      user.pastemode=false;
      return true;
    }
    if (label.equals("Invert")) {
      user.undobuf=new IconBuf(user.tilepanel.pixels,
			       user.tilepanel.WIDTH,
			       user.tilepanel.HEIGHT,
			       user.captx1,user.capty1,
			       user.captx2,user.capty2);
      user.tilepanel.invertRect(user.captx1,user.capty1,
			      user.captx2,user.capty2);
      user.pastemode=false;
      user.tilepanel.resync();
      user.tileicon.putIcon();
      return true;
    }
    if (label.equals("Flip Horiz.") || label.equals("Flip Vert.")) {
      user.undobuf=new IconBuf(user.tilepanel.pixels,
			       user.tilepanel.WIDTH,
			       user.tilepanel.HEIGHT,
			       user.captx1,user.capty1,
			       user.captx2,user.capty2);
      user.tilepanel.flipRect(user.captx1,user.capty1,
			      user.captx2,user.capty2,
			      label.equals("Flip Horiz.") ? true : false);
      user.pastemode=false;
      user.tilepanel.resync();
      user.tileicon.putIcon();
      return true;
    }
    /*
    if (label.equals("Rotate 90") || label.equals("Rotate -90")) {
      user.undobuf=new IconBuf(user.tilepanel.pixels,
			       user.tilepanel.WIDTH,
			       user.tilepanel.HEIGHT,
			       user.captx1,user.capty1,
			       user.captx2,user.capty2);
      user.tilepanel.rotateRect(user.tilepanel.osGraphics,
			      user.captx1,user.capty1,
			      user.captx2,user.capty2,
			      label.equals("Rotate 90") ? true : false);
      user.pastemode=false;
      user.tilepanel.repaint();
      user.tileicon.repaint();
      return true;
    }
    */
    if (label.equals("Resize")) {
      showOptionDialog(false);
      return true;
    }
    if (label.equals("Clear Icon")) {
      user.clearIcon();
      user.pastemode=false;
      user.tilepanel.resync();
      user.tileicon.putIcon();
      return true;
    }
    if (label.equals("x2")) {
      user.magni=2;
      user.reset();
      return true;
    }
    if (label.equals("x4")) {
      user.magni=4;
      user.reset();
      return true;
    }
    if (label.equals("x6")) {
      user.magni=6;
      user.reset();
      return true;
    }
    if (label.equals("x8")) {
      user.magni=8;
      user.reset();
      return true;
    }
    if (label.equals("x10")) {
      user.magni=10;
      user.reset();
      return true;
    }
    return super.handleEvent(evt);
  }
  if (evt.id == Event.WINDOW_DESTROY) {
    if (inAnApplet) {
      dispose();
      return true;
    } else {
      System.exit(0);
    }
  }
  return super.handleEvent(evt);
}
  

public Dimension preferredSize() {
  return(new Dimension(500,500));
}

public Dimension minimumSize() {
  return(new Dimension(500,500));
}

public static void main(String args[]) {
  inAnApplet=false;
  bitedit window = new bitedit();
  window.reshape(100,100,500,500);
  window.show();  
}

}
